﻿
''' <summary>
''' 
''' A simple widget that can be resized
''' 
''' Credits to Jesse Chunn for his cool Resizable Control code...
''' http://www.codeproject.com/Articles/20716/Allow-the-User-to-Resize-Controls-at-Runtime
''' 
''' </summary>
''' <remarks></remarks>

Public Class Widget

    Inherits System.Windows.Forms.UserControl

    Private Enum EdgeEnum
        None
        Right
        Left
        Top
        Bottom
        TopLeft
    End Enum

    Public mType As Integer
    Public mLeft As Integer
    Public mTop As Integer
    Public mWidth As Integer
    Public mHeight As Integer
    Public mName As String
    Public mCaption As String
    Public mTextAlign As Integer

    Public mEnabled As Integer
    Public mHidden As Integer
    Public mBorderType As Integer
    Public mAlpha As String

    Public mColor As String
    Public mColorARGB(3) As Integer

    Public mTextColor As String
    Public mTextColorARGB(3) As Integer

    Public mToggleState As Integer
    Public mToggleGroup As String
    Public mOnStateChanged As String

    Public mMin As String
    Public mMax As String
    Public mSteps As String
    Public mValue As String
    Public mOnValueChanged As String
    Public mOrientation As Integer

    Public mVisibleLines As Integer
    Public mList As String
    Public mOnIndexChanged As String
    Public mOnItemRemoved As String

    Public mInputType As Integer
    Public mOnTextChanged As String

    Public mRows As Integer

    Public mOnPress As String
    Public mOnRelease As String
    Public mOnDrag As String

    Public mOnOpen As String
    Public mOnClose As String
    Public mHasControls As Integer

    Public mParentGroup As String = ""
    Public mZIndex As Single

    Public mImageName As String
    Public mScaleX As String
    Public mScaleY As String
    Public mImageRotation As Single

    ' MARKED AS WIDGET VARS BUT DONT NEED TO BE PERSISTENT
    Public mCanResizeX As Boolean
    Public mCanResizeY As Boolean

    Public mHandleX As Integer
    Public mHandleY As Integer

    Public mImage As Image
    Public mBitmap As Bitmap

    ' FOLLOWING ARE NOT ACTUAL VARIABLES IN A WIDGET
    Private mMouseDown As Boolean = False
    Private mEdge As EdgeEnum = EdgeEnum.None
    Private mOutlineDrawn As Boolean = False
    Private bWidth As Integer = 4
    Private wSize As Integer = 48

    Private mMaxX As Integer = 0
    Private mMaxY As Integer = 0
    Private mMinX As Integer = 0
    Private mMinY As Integer = 0

    Private Function clamp(ByVal value As Integer, ByVal min As Integer, ByVal max As Integer) As Integer
        If value < min Then
            Return min
        ElseIf value >= max Then
            Return max
        End If
        Return value
    End Function

    Protected Overrides ReadOnly Property CreateParams() As System.Windows.Forms.CreateParams
        Get
            Dim cp As CreateParams = MyBase.CreateParams
            cp.ExStyle = cp.ExStyle Or &H20  ' Turn on WS_EX_TRANSPARENT
            Return cp
        End Get
    End Property

    Protected Overrides Sub OnPaintBackground(ByVal e As PaintEventArgs)
        MyBase.OnPaintBackground(e)
    End Sub

    Public Overrides Property BackColor() As Color
        Get
            Return MyBase.BackColor
        End Get
        Set(ByVal value As Color)
            MyBase.BackColor = Color.Transparent
        End Set
    End Property

    Public Sub New(ByVal name As String, ByVal type As Integer)
        SetStyle(ControlStyles.AllPaintingInWmPaint, True)
        SetStyle(ControlStyles.UserPaint, True)
        SetStyle(ControlStyles.DoubleBuffer, True)

        mType = type

        mName = name.ToUpper & "_" & UID.ToString("D4")
        mLeft = 48
        mTop = 48
        mHeight = 48
        mWidth = 48

        name = mName.ToUpper & "_" & UID.ToString("D4")
        Left = mLeft
        Top = mTop
        Width = mWidth
        Height = mHeight
        mCaption = ""
        mHasControls = 0
        mTextAlign = textAlign.MIDDLE
        mAlpha = ""
        mTextColor = "EAEBEC"
        mTextColorARGB(0) = 255
        mTextColorARGB(1) = 234
        mTextColorARGB(2) = 235
        mTextColorARGB(3) = 236
        mColor = "C0C0C0"
        mColorARGB(0) = 255
        mColorARGB(1) = 192
        mColorARGB(2) = 192
        mColorARGB(3) = 192
        mBorderType = borderType.RAISED
        mOrientation = 0
        mEnabled = 1
        mHidden = 0

        mToggleState = 0
        mToggleGroup = ""
        mOnStateChanged = ""

        mMin = ""
        mMax = ""
        mSteps = ""
        mValue = ""
        mOnValueChanged = ""

        mList = ""
        mInputType = 0
        mRows = 3
        mVisibleLines = 3
        mOnTextChanged = ""
        mOnPress = ""
        mOnRelease = ""
        mOnDrag = ""
        mOnIndexChanged = ""
        mOnItemRemoved = ""
        mOnOpen = ""
        mOnClose = ""
        mParentGroup = ""
        mZIndex = 0.5
        mScaleX = "1.0"
        mScaleY = "1.0"
        mImageName = ""

        mImage = Nothing
        mBitmap = Nothing

        'mNumChildren = 0
        ' BY DEFAULT ALL WIDGETS CAN BE RESIZED ON BOTH AXIS
        mCanResizeX = True
        mCanResizeY = True

        ' SET SPECIFIC PROPERTIES DEPENDING ON WIDGET TYPE
        Select Case mType
            Case widgetType.WINDOW
                mBorderType = borderType.WINDOW
                Width = 640
                Height = 480
                mTextAlign = textAlign.TOP_LEFT
                mZIndex = 0
                MinimumSize = New Size(wSize * 2, wSize * 2)

            Case widgetType.IMAGE
                mBorderType = borderType.NONE
                mCanResizeX = False
                mCanResizeY = False
                mZIndex = 0.3

            Case widgetType.PANEL
                mBorderType = borderType.RECESSED
                Width = wSize * 2
                Height = wSize * 2
                mZIndex = 0.1
                MinimumSize = New Size(wSize * 2, wSize * 2)

            Case widgetType.BUTTON
                Width = wSize * 3
                mBorderType = borderType.BUTTON
                mCanResizeY = False
                mTextAlign = textAlign.MIDDLE
                MinimumSize = New Size(wSize * 2, wSize * 1)

            Case widgetType.LABEL
                mCanResizeY = False
                mBorderType = borderType.LABEL
                Width = wSize * 2
                mTextAlign = textAlign.MIDDLE
                MinimumSize = New Size(wSize * 2, wSize * 1)

            Case widgetType.LIST
                mBorderType = borderType.RECESSED
                Width = wSize * 4
                Height = wSize * 3
                MinimumSize = New Size(wSize * 2, wSize * 2)

            Case widgetType.LOG
                mBorderType = borderType.RECESSED
                Width = wSize * 4
                Height = wSize * 2
                MinimumSize = New Size(wSize * 2, wSize * 2)

            Case widgetType.CHECKBOX
                mBorderType = borderType.CHECKBOX
                mCanResizeX = False
                mCanResizeY = False
                mTextAlign = textAlign.LEFT
                mCaption = "CheckBox"

            Case widgetType.RADIOBUTTON
                mBorderType = borderType.RADIOBUTTON
                mCanResizeX = False
                mCanResizeY = False
                mTextAlign = textAlign.LEFT
                mCaption = "RadioButton"

            Case widgetType.TEXTVIEW
                mBorderType = borderType.RECESSED
                Width = wSize * 4
                Height = wSize * 3
                MinimumSize = New Size(wSize * 3, wSize * 2)

            Case widgetType.TEXTINPUT
                mBorderType = borderType.RECESSED
                mCanResizeY = False
                Width = wSize * 3
                mTextAlign = textAlign.MIDDLE_LEFT
                MinimumSize = New Size(wSize * 2, wSize * 1)

            Case widgetType.HSCROLLBAR
                mBorderType = borderType.HSCROLLBAR
                mCanResizeY = False
                Width = wSize * 3
                mOrientation = 0
                mMin = "0"
                mMax = "0"
                mSteps = "0"
                mValue = "0"
                MinimumSize = New Size(wSize * 3, wSize * 1)

            Case widgetType.VSCROLLBAR
                mBorderType = borderType.VSCROLLBAR
                mCanResizeX = False
                Height = wSize * 3
                mOrientation = 1
                mMin = "0"
                mMax = "0"
                mSteps = "0"
                mValue = "0"
                MinimumSize = New Size(wSize * 1, wSize * 3)

            Case Else
        End Select

        'Invalidate()

    End Sub

#Region "Resizable Control Code: widget__MouseDown, widget_MouseUp, widget_MouseMove, widget_MouseLeave"

    Private Sub widget__MouseDown(ByVal sender As Object, ByVal e As System.Windows.Forms.MouseEventArgs) Handles Me.MouseDown
        If e.Button = Windows.Forms.MouseButtons.Left Then
            mMouseDown = True

            Dim localMousePosition As Point = Form1.PointToClient(Cursor.Position)
            mLeft = localMousePosition.X
            mTop = localMousePosition.Y


            ' GET POSITION OF MOUSE INSIDE CONTROL
            Dim localMouse As Point = PointToClient(Cursor.Position)
            mHandleX = localMouse.X '- Left
            mHandleY = localMouse.Y '- Top

            mMaxX = (Me.Left + Me.Width) - wSize
            mMaxY = (Me.Top + Me.Height) - wSize

            'Dim g As Graphics = Me.CreateGraphics
            'Dim rect As New Rectangle(0, 0, Me.Width, Me.Height)
            'Dim pen As New Pen(Color.GreenYellow, 4)
            'g.DrawRectangle(pen, rect)

            Form1.selectWidget(CType(sender, Widget)) ' BAD CODE :D
        End If
    End Sub

    Private Sub widget_MouseUp(ByVal sender As Object, ByVal e As System.Windows.Forms.MouseEventArgs) Handles Me.MouseUp
        mMouseDown = False
    End Sub

    Private Sub widget_MouseLeave(ByVal sender As Object, ByVal e As System.EventArgs) Handles Me.MouseLeave
        mEdge = EdgeEnum.None
        Me.Refresh()
    End Sub

    Private Sub widget_MouseMove(ByVal sender As Object, ByVal e As System.Windows.Forms.MouseEventArgs) Handles Me.MouseMove

        Dim g As Graphics = Me.CreateGraphics

        Dim localMousePosition As Point = Form1.PointToClient(Cursor.Position) ' ANTIX

        Dim dx As Integer = -(mLeft - localMousePosition.X)
        Dim dy As Integer = -(mTop - localMousePosition.Y)
        mLeft = localMousePosition.X
        mTop = localMousePosition.Y

        Select Case mEdge
            Case EdgeEnum.Left
                g.FillRectangle(Brushes.Fuchsia, 0, 0, bWidth, Me.Height)
                mOutlineDrawn = True
            Case EdgeEnum.Right
                g.FillRectangle(Brushes.Fuchsia, Me.Width - bWidth, 0, Me.Width, Me.Height)
                mOutlineDrawn = True
            Case EdgeEnum.Top
                g.FillRectangle(Brushes.Fuchsia, 0, 0, Me.Width, bWidth)
                mOutlineDrawn = True
            Case EdgeEnum.Bottom
                g.FillRectangle(Brushes.Fuchsia, 0, Me.Height - bWidth, Me.Width, bWidth)
                mOutlineDrawn = True
            Case EdgeEnum.None
                Dim rect As New Rectangle(0, 0, Me.Width, Me.Height)
                Dim pen As New Pen(Color.GreenYellow, 4)
                g.DrawRectangle(pen, rect)

                If mOutlineDrawn Then
                    Me.Refresh()
                    mOutlineDrawn = False
                End If
        End Select

        If mMouseDown Then
            Me.SuspendLayout()
            Dim parentWidth As Integer = Me.Parent.Width
            Dim parentHeight As Integer = Me.Parent.Height

            Select Case mEdge
                Case EdgeEnum.Left
                    If mCanResizeX Then
                        Dim left As Integer = clamp(Me.Left + e.X, 0, mMaxX)
                        Dim width As Integer = clamp(Me.Width - e.X, wSize, mMaxX + wSize)
                        Me.SetBounds(left, Me.Top, width, Me.Height)
                    End If
                Case EdgeEnum.Right
                    If mCanResizeX Then
                        Dim width As Integer = clamp(Me.Width - (Me.Width - e.X), wSize, parentWidth - Me.Left)
                        Me.SetBounds(Me.Left, Me.Top, width, Me.Height)
                    End If
                Case EdgeEnum.Top
                    If mCanResizeY Then
                        Dim top As Integer = clamp(Me.Top + e.Y, 0, mMaxY)
                        Dim height As Integer = clamp(Me.Height - e.Y, wSize, mMaxY + wSize)
                        Me.SetBounds(Me.Left, top, Me.Width, height)
                    End If
                Case EdgeEnum.Bottom
                    If mCanResizeY Then
                        Dim height As Integer = clamp(Me.Height - (Me.Height - e.Y), wSize, parentHeight - Me.Top)
                        Me.SetBounds(Me.Left, Me.Top, Me.Width, height)
                    End If
                Case EdgeEnum.None

                    If snapToGrid Then
                        Dim localMouse As Point = Parent.PointToClient(Cursor.Position)
                        If Not shiftHeld Then
                            Dim left As Integer = CInt(((localMouse.X - mHandleX) \ gridSizeX) * gridSizeX)
                            Me.Left = clamp(left, 0, parentWidth - Me.Width)
                        End If
                        If Not ctrlHeld Then
                            Dim top As Integer = CInt(((localMouse.Y - mHandleY) \ gridSizeY) * gridSizeY)
                            Me.Top = clamp(top, 0, parentHeight - Me.Height)
                        End If
                    Else
                        If Not shiftHeld Then
                            Me.Left = clamp(Me.Left + dx, 0, parentWidth - Me.Width)
                        End If
                        If Not ctrlHeld Then
                            Me.Top = clamp(Me.Top + dy, 0, parentHeight - Me.Height)
                        End If
                    End If
            End Select

            Me.Left = clamp(Me.Left, 0, parentWidth)
            Me.Top = clamp(Me.Top, 0, parentHeight)

            Form1.widgetLeft.Value = Me.Left ' THESE COULD COME BACK TO HAUNT ME LATER
            Form1.widgetTop.Value = Me.Top
            Form1.widgetWidth.Value = Me.Width
            Form1.widgetHeight.Value = Me.Height

            Me.ResumeLayout()

            ' MOUSEBUTTON NOT HELD DOWN SO DO HOVERY STUFF
        Else
            Select Case True
                Case e.X <= bWidth 'LEFT EDGE
                    Me.Cursor = Cursors.SizeWE
                    mEdge = EdgeEnum.Left
                Case e.X > Me.Width - (bWidth + 1) 'RIGHT EDGE
                    Me.Cursor = Cursors.SizeWE
                    mEdge = EdgeEnum.Right
                Case e.Y <= bWidth 'top edge
                    Me.Cursor = Cursors.SizeNS
                    mEdge = EdgeEnum.Top
                Case e.Y > Me.Height - (bWidth + 1) 'BOTTOM EDGE
                    Me.Cursor = Cursors.SizeNS
                    mEdge = EdgeEnum.Bottom
                Case Else 'no edge
                    Me.Cursor = Cursors.SizeAll '.DEFAULT
                    mEdge = EdgeEnum.None
            End Select
        End If
    End Sub
#End Region

    ' DRAW WIDGET BORDER
    Private Sub drawBorder(ByVal w As Widget, ByVal g As Graphics)
        If w.mBorderType <> borderType.NONE Then ' ONLY DRAW IF THERE IS ACTUALLY A BORDER SET

            Dim l As Integer = w.Left
            Dim t As Integer = w.Top
            If w.mType = widgetType.WINDOW Then
                l = 0
                t = 0
            End If

            Dim bX As Integer = 0 ' BORDERTYPE.WINDOW IS AT 0,0
            Dim bY As Integer = 0
            Select Case w.mBorderType
                Case borderType.BUTTON
                    bX = 192
                    bY = 0
                Case borderType.RAISED
                    bX = 0
                    bY = 192
                Case borderType.RECESSED
                    bX = 192
                    bY = 192
                Case borderType.HSCROLLBAR
                    bX = -1 * wSize
                    bY = wSize * 8
                Case borderType.VSCROLLBAR
                    bX = wSize * 8
                    bY = wSize * -1
            End Select

            Dim sRect As New Rectangle(bX, bY, wSize, wSize)
            Dim dRect As New Rectangle(l, t, wSize, wSize)

            ' DRAW SINGLE IMAGE (SPECIAL CASES FOR RADIOBUTTONS AND CHECKBOXES)
            If w.Width = wSize And w.Height = wSize Then
                Select Case w.mType
                    Case widgetType.CHECKBOX
                        If w.mToggleState = 0 Then
                            sRect.X = wSize * 0 ' UNCHECKED
                            sRect.Y = wSize * 9
                        Else
                            sRect.X = wSize * 1 ' CHECKED
                            sRect.Y = wSize * 9
                        End If
                    Case widgetType.RADIOBUTTON
                        If w.mToggleState = 0 Then
                            sRect.X = wSize * 2 ' UNCHECKED
                            sRect.Y = wSize * 9
                        Else
                            sRect.X = wSize * 3 ' CHECKED
                            sRect.Y = wSize * 9
                        End If
                End Select
                g.DrawImage(atlasBitmap, dRect, sRect, GraphicsUnit.Pixel)

            ElseIf w.Height = wSize Then
                ' DRAW HORIZONTAL WIDGET, 1 TILE HIGH
                sRect.X = bX + wSize
                g.DrawImage(atlasBitmap, dRect, sRect, GraphicsUnit.Pixel)

                sRect.X = bX + wSize * 3
                dRect.X = l + w.Width - wSize
                g.DrawImage(atlasBitmap, dRect, sRect, GraphicsUnit.Pixel)

                sRect.X = bX + wSize * 2
                dRect.X = l + wSize
                dRect.Width = w.Width - wSize * 2
                g.DrawImage(atlasBitmap, dRect, sRect, GraphicsUnit.Pixel)

                If w.mType = widgetType.HSCROLLBAR Then
                    sRect.X = wSize * 9
                    sRect.Y = wSize * 0
                    dRect.X = l
                    sRect.Width = wSize * 1
                    dRect.Width = wSize * 1
                    g.DrawImage(atlasBitmap, dRect, sRect, GraphicsUnit.Pixel)
                End If

            ElseIf w.Width = wSize Then
                ' DRAW VERTICAL WIDGET, 1 TILE HIGH
                sRect.Y = bY + wSize
                g.DrawImage(atlasBitmap, dRect, sRect, GraphicsUnit.Pixel)

                sRect.Y = bY + wSize * 3
                dRect.Y = t + w.Height - wSize
                g.DrawImage(atlasBitmap, dRect, sRect, GraphicsUnit.Pixel)

                sRect.Y = bY + wSize * 2
                dRect.Y = t + wSize
                dRect.Height = w.Height - wSize * 2
                g.DrawImage(atlasBitmap, dRect, sRect, GraphicsUnit.Pixel)

                If w.mType = widgetType.VSCROLLBAR Then
                    sRect.X = wSize * 9
                    sRect.Y = wSize * 1
                    dRect.Y = t
                    sRect.Height = wSize * 1
                    dRect.Height = wSize * 1
                    g.DrawImage(atlasBitmap, dRect, sRect, GraphicsUnit.Pixel)
                End If
            Else
                ' DRAW CORNERS
                sRect.X = bX + wSize * 1 ' TOP LEFT
                sRect.Y = bY + wSize * 1
                g.DrawImage(atlasBitmap, dRect, sRect, GraphicsUnit.Pixel)

                sRect.X = bX + wSize * 3 ' TOP RIGHT
                sRect.Y = bY + wSize * 1
                dRect.X = l + w.Width - wSize
                g.DrawImage(atlasBitmap, dRect, sRect, GraphicsUnit.Pixel)

                sRect.X = bX + wSize * 1 ' BOTTOM LEFT
                sRect.Y = bY + wSize * 3
                dRect.X = l + 0
                dRect.Y = t + w.Height - wSize
                g.DrawImage(atlasBitmap, dRect, sRect, GraphicsUnit.Pixel)

                sRect.X = bX + wSize * 3 ' BOTTOM RIGHT
                dRect.X = l + w.Width - wSize
                dRect.Y = t + w.Height - wSize
                g.DrawImage(atlasBitmap, dRect, sRect, GraphicsUnit.Pixel)

                ' DRAW LEFT AND RIGHT BORDERS
                If w.Width >= wSize * 2 Then
                    sRect.X = bX + wSize * 1 ' LEFT
                    sRect.Y = bY + wSize * 2
                    dRect.X = l + 0
                    dRect.Y = t + wSize
                    dRect.Height = w.Height - wSize * 2
                    g.DrawImage(atlasBitmap, dRect, sRect, GraphicsUnit.Pixel)
                    sRect.X = bX + wSize * 3 ' RIGHT
                    dRect.X = l + w.Width - wSize
                    g.DrawImage(atlasBitmap, dRect, sRect, GraphicsUnit.Pixel)
                End If

                ' DRAW TOP AND BOTTOM BORDERS
                If w.Height >= wSize * 2 Then
                    sRect.X = bX + wSize * 2 ' TOP
                    sRect.Y = bY + wSize * 1
                    dRect.X = l + wSize
                    dRect.Y = t + 0
                    dRect.Width = w.Width - wSize * 2
                    dRect.Height = wSize
                    g.DrawImage(atlasBitmap, dRect, sRect, GraphicsUnit.Pixel)
                    sRect.Y = bY + wSize * 3 ' BOTTOM
                    dRect.Y = t + w.Height - wSize
                    g.DrawImage(atlasBitmap, dRect, sRect, GraphicsUnit.Pixel)
                End If

                ' DRAW MIDDLE BIT
                If w.Width >= wSize * 2 And w.Height >= wSize * 2 Then
                    sRect.X = bX + wSize * 2 ' MIDDLE
                    sRect.Y = bY + wSize * 2
                    dRect.X = l + wSize
                    dRect.Y = t + wSize
                    dRect.Width = w.Width - wSize * 2
                    dRect.Height = w.Height - wSize * 2
                    g.DrawImage(atlasBitmap, dRect, sRect, GraphicsUnit.Pixel)
                End If
            End If

        End If
    End Sub

    ' DRAW WIDGET CAPTION
    Private Sub drawCaption(ByVal w As Widget, ByVal g As Graphics)
        If w.mCaption <> "" Then ' ONLY DRAW IF THERE IS ACTUALLY A CAPTION

            Dim l As Integer = w.Left
            Dim t As Integer = w.Top
            If w.mType = widgetType.WINDOW Then
                l = 0
                t = 0
            End If

            Dim brush As New SolidBrush(Color.FromArgb(w.mTextColorARGB(0), w.mTextColorARGB(1), w.mTextColorARGB(2), w.mTextColorARGB(3)))


            Dim ss As New SizeF
            ss = g.MeasureString(w.mCaption, w.Font)
            Dim x, y As Single
            Select Case w.mTextAlign
                Case textAlign.TOP_LEFT
                    x = l + 5
                    y = t + 5
                Case textAlign.TOP_CENTER
                    x = CSng(l + w.Width / 2 - ss.Width / 2)
                    y = t + 5
                Case textAlign.TOP_RIGHT
                    x = l + (w.Width - 5 - ss.Width)
                    y = t + 5
                Case textAlign.MIDDLE_LEFT
                    x = l + 5
                    y = CSng(t + w.Height / 2 - ss.Height / 2)
                Case textAlign.MIDDLE
                    x = CSng(l + w.Width / 2 - ss.Width / 2)
                    y = CSng(t + w.Height / 2 - ss.Height / 2)
                Case textAlign.MIDDLE_RIGHT
                    x = l + (w.Width - 5 - ss.Width)
                    y = CSng(t + w.Height / 2 - ss.Height / 2)
                Case textAlign.BOTTOM_LEFT
                    x = l + 5
                    y = CSng(t + w.Height - 5 - ss.Height)
                Case textAlign.MIDDLE_BOTTOM
                    x = CSng(l + w.Width / 2 - ss.Width / 2)
                    y = CSng(t + w.Height - 5 - ss.Height)
                Case textAlign.BOTTOM_RIGHT
                    x = l + (w.Width - 5 - ss.Width)
                    y = CSng(t + w.Height - 5 - ss.Height)
                Case textAlign.LEFT
                    x = l - 5 - ss.Width
                    y = CSng(t + w.Height / 2 - ss.Height / 2)
                Case textAlign.RIGHT
                    x = l + w.Width + 5
                    y = CSng(t + w.Height / 2 - ss.Height / 2)
                Case textAlign.ABOVE
                    x = CSng(l + w.Width / 2 - ss.Width / 2)
                    y = t - 5 - ss.Height
                Case textAlign.BELOW
                    x = CSng(l + w.Width / 2 - ss.Width / 2)
                    y = t + w.Height + 5
            End Select
            g.DrawString(w.mCaption, Me.Font, brush, x, y) ' DRAW IT
        End If
    End Sub

    Public Class sortedItems
        Public index As Integer
        Public zOrder As Single
        Public Sub New(ByVal i As Integer, ByVal z As Single)
            index = i
            zOrder = z
        End Sub
    End Class

    ' REPAINT WIDGET
    Protected Overrides Sub OnPaint(ByVal e As PaintEventArgs)
        If mType = widgetType.WINDOW Then ' ONLY REPAINT WHEN A WINDOW IS INVALIDATED!!
            Dim g As Graphics = e.Graphics
            drawBorder(Me, g)
            drawCaption(Me, g)

            ' CREATE A LIST THAT WILL BE SORTED BY WIDGET ZORDER
            Dim sList As New List(Of sortedItems)
            Dim i As Integer = 0
            For Each c As Control In Controls
                If TypeOf c Is Widget Then
                    Dim w As Widget = CType(c, Widget)
                    sList.Add(New sortedItems(i, w.mZIndex)) ' JUST ADD THEM ALL TO THE LIST INITIALLY
                    i = i + 1
                End If
            Next
            sList.Sort(Function(x, y) x.zOrder.CompareTo(y.zOrder)) ' SORT BY ZORDER

            For i = 0 To sList.Count - 1
                Dim w As Widget = CType(Controls(sList(i).index), Widget) ' NEXT WIDGET TO BE DRAWN

                If w.mType = widgetType.IMAGE Then
                    If w.mBitmap IsNot Nothing Then
                        ' DRAW IMAGE
                        Dim sr As New Rectangle(0, 0, w.Width, w.Height)
                        Dim dr As New Rectangle(w.Left, w.Top, w.Width, w.Height)
                        g.DrawImage(w.mBitmap, dr, sr, GraphicsUnit.Pixel)
                    End If
                Else
                    ' DRAW NORMAL BORDER AND CAPTION
                    drawBorder(w, g)
                    drawCaption(w, g)
                End If
            Next
        End If
    End Sub

End Class
